# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from http.client import HTTPConnection
from collections import OrderedDict
import logging
import pytest
from unittest import mock
from socket import timeout as SocketTimeout

from urllib3.util import connection as urllib_connection
from urllib3.response import HTTPResponse as UrllibResponse
from urllib3.connection import HTTPConnection as UrllibConnection

from azure.core.rest._http_response_impl import HttpResponseImpl as RestHttpResponseImpl
from azure.core.pipeline._tools import is_rest
from azure.core.pipeline.transport import HttpResponse as PipelineTransportHttpResponse, RequestsTransport
from azure.core.pipeline.transport._base import HttpTransport, _deserialize_response, _urljoin
from azure.core.pipeline.policies import HeadersPolicy
from azure.core.pipeline import Pipeline
from azure.core.exceptions import (
    HttpResponseError,
    ServiceRequestError,
    ServiceResponseError,
    ServiceRequestTimeoutError,
    ServiceResponseTimeoutError,
)

from utils import (
    HTTP_REQUESTS,
    request_and_responses_product,
    HTTP_CLIENT_TRANSPORT_RESPONSES,
    create_transport_response,
)


class PipelineTransportMockResponse(PipelineTransportHttpResponse):
    def __init__(self, request, body, content_type):
        super(PipelineTransportMockResponse, self).__init__(request, None)
        self._body = body
        self.content_type = content_type

    def body(self):
        return self._body


class RestMockResponse(RestHttpResponseImpl):
    def __init__(self, request, body, content_type):
        super(RestMockResponse, self).__init__(
            request=request,
            internal_response=None,
            content_type=content_type,
            block_size=None,
            status_code=200,
            reason="OK",
            headers={},
            stream_download_generator=None,
        )
        # the impl takes in a lot more kwargs. It's not public and is a
        # helper implementation shared across our azure core transport responses
        self._content = body
        self._is_closed = True
        self._is_stream_consumed = True


MOCK_RESPONSES = [PipelineTransportMockResponse, RestMockResponse]


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_http_request_serialization(http_request):
    # Method + Url
    request = http_request("DELETE", "/container0/blob0")
    serialized = request.serialize()

    expected = (
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        # No headers
        b"\r\n"
    )
    assert serialized == expected

    # Method + Url + Headers
    request = http_request(
        "DELETE",
        "/container0/blob0",
        # Use OrderedDict to get consistent test result on 3.5 where order is not guaranteed
        headers=OrderedDict(
            {
                "x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT",
                "Authorization": "SharedKey account:G4jjBXA7LI/RnWKIOQ8i9xH4p76pAQ+4Fs4R1VxasaE=",  # fake key suppressed in credscan
                "Content-Length": "0",
            }
        ),
    )
    serialized = request.serialize()

    expected = (
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"Authorization: SharedKey account:G4jjBXA7LI/RnWKIOQ8i9xH4p76pAQ+4Fs4R1VxasaE=\r\n"  # fake key suppressed in credscan
        b"Content-Length: 0\r\n"
        b"\r\n"
    )
    assert serialized == expected

    # Method + Url + Headers + Body
    request = http_request(
        "DELETE",
        "/container0/blob0",
        headers={
            "x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT",
        },
    )
    request.set_bytes_body(b"I am groot")
    serialized = request.serialize()

    expected = (
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"Content-Length: 10\r\n"
        b"\r\n"
        b"I am groot"
    )
    assert serialized == expected


def test_url_join():
    assert _urljoin("devstoreaccount1", "?testdir") == "devstoreaccount1?testdir"
    assert _urljoin("devstoreaccount1", "?testdir=foo") == "devstoreaccount1?testdir=foo"
    assert _urljoin("devstoreaccount1/api", "?a=1") == "devstoreaccount1/api?a=1"
    assert (
        _urljoin("devstoreaccount1", "/?restype=service&comp=properties")
        == "devstoreaccount1/?restype=service&comp=properties"
    )
    assert _urljoin("devstoreaccount1", "") == "devstoreaccount1"
    assert _urljoin("devstoreaccount1", "testdir/") == "devstoreaccount1/testdir/"
    assert _urljoin("devstoreaccount1/", "") == "devstoreaccount1/"
    assert _urljoin("devstoreaccount1/", "/testdir/") == "devstoreaccount1/testdir/"
    assert _urljoin("devstoreaccount1/", "testdir/") == "devstoreaccount1/testdir/"
    assert _urljoin("devstoreaccount1?a=1", "testdir/") == "devstoreaccount1/testdir/?a=1"
    assert _urljoin("devstoreaccount1", "testdir/?b=2") == "devstoreaccount1/testdir/?b=2"
    assert _urljoin("devstoreaccount1?a=1", "testdir/?b=2") == "devstoreaccount1/testdir/?a=1&b=2"
    assert _urljoin("devstoreaccount1", "documentModels:build") == "devstoreaccount1/documentModels:build"


@pytest.mark.parametrize("http_request,http_response", request_and_responses_product(HTTP_CLIENT_TRANSPORT_RESPONSES))
def test_http_client_response(port, http_request, http_response):
    # Create a core request
    request = http_request("GET", "http://localhost:{}".format(port))

    # Fake a transport based on http.client
    conn = HTTPConnection("localhost", port)
    conn.request("GET", "/get")
    r1 = conn.getresponse()

    response = create_transport_response(http_response, request, r1)
    if is_rest(http_response):
        response.read()

    # Don't assume too much in those assert, since we reach a real server
    assert response.internal_response is r1
    assert response.reason is not None
    assert isinstance(response.status_code, int)
    assert len(response.headers.keys()) != 0
    assert len(response.text()) != 0
    assert "content-type" in response.headers
    assert "Content-Type" in response.headers


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_response_deserialization(http_request):

    # Method + Url
    request = http_request("DELETE", "/container0/blob0")
    body = (
        b"HTTP/1.1 202 Accepted\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
    )

    response = _deserialize_response(body, request)

    assert response.status_code == 202
    assert response.reason == "Accepted"
    assert response.headers == {"x-ms-request-id": "778fdc83-801e-0000-62ff-0334671e284f", "x-ms-version": "2018-11-09"}

    # Method + Url + Headers + Body
    request = http_request(
        "DELETE",
        "/container0/blob0",
        headers={
            "x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT",
        },
    )
    request.set_bytes_body(b"I am groot")
    body = (
        b"HTTP/1.1 200 OK\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"I am groot"
    )

    response = _deserialize_response(body, request)

    assert isinstance(response.status_code, int)
    assert response.reason == "OK"
    assert response.headers == {"x-ms-request-id": "778fdc83-801e-0000-62ff-0334671e284f", "x-ms-version": "2018-11-09"}
    assert response.text() == "I am groot"


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_response_deserialization_utf8_bom(http_request):

    request = http_request("DELETE", "/container0/blob0")
    body = (
        b"HTTP/1.1 400 One of the request inputs is not valid.\r\n"
        b"x-ms-error-code: InvalidInput\r\n"
        b"x-ms-request-id: 5f3f9f2f-e01e-00cc-6eb1-6d00b5000000\r\n"
        b"x-ms-version: 2019-02-02\r\n"
        b"Content-Length: 220\r\n"
        b"Content-Type: application/xml\r\n"
        b"Server: Windows-Azure-Blob/1.0\r\n"
        b"\r\n"
        b'\xef\xbb\xbf<?xml version="1.0" encoding="utf-8"?>\n<Error><Code>InvalidInput</Code><Message>One'
        b"of the request inputs is not valid.\nRequestId:5f3f9f2f-e01e-00cc-6eb1-6d00b5000000\nTime:2019-09-17T23:44:07.4671860Z</Message></Error>"
    )
    response = _deserialize_response(body, request)
    assert response.body().startswith(b"\xef\xbb\xbf")


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send(http_request):

    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    req0 = http_request("DELETE", "/container0/blob0")
    req1 = http_request("DELETE", "/container1/blob1")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        req0,
        req1,
        policies=[header_policy],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",  # Fix it so test are deterministic
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send_with_context(http_request):
    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    req0 = http_request("DELETE", "/container0/blob0")
    req1 = http_request("DELETE", "/container1/blob1")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        req0,
        req1,
        policies=[header_policy],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",  # Fix it so test are deterministic
        headers={"Accept": "application/json"},
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"Accept: application/json\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"Accept: application/json\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send_with_one_changeset(http_request):

    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    requests = [http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")]

    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        *requests, policies=[header_policy], boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        changeset,
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send_with_multiple_changesets(http_request):

    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    changeset1 = http_request("", "")
    changeset1.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        http_request("DELETE", "/container1/blob1"),
        policies=[header_policy],
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    changeset2 = http_request("", "")
    changeset2.set_multipart_mixed(
        http_request("DELETE", "/container2/blob2"),
        http_request("DELETE", "/container3/blob3"),
        policies=[header_policy],
        boundary="changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314",
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        changeset1,
        changeset2,
        policies=[header_policy],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 3\r\n"
        b"\r\n"
        b"DELETE /container3/blob3 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send_with_combination_changeset_first(http_request):

    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        http_request("DELETE", "/container1/blob1"),
        policies=[header_policy],
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        changeset,
        http_request("DELETE", "/container2/blob2"),
        policies=[header_policy],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send_with_combination_changeset_last(http_request):

    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container1/blob1"),
        http_request("DELETE", "/container2/blob2"),
        policies=[header_policy],
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        changeset,
        policies=[header_policy],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_multipart_send_with_combination_changeset_middle(http_request):

    transport = mock.MagicMock(spec=HttpTransport)

    header_policy = HeadersPolicy({"x-ms-date": "Thu, 14 Jun 2018 16:46:54 GMT"})

    changeset = http_request("", "")
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container1/blob1"),
        policies=[header_policy],
        boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"),
        changeset,
        http_request("DELETE", "/container2/blob2"),
        policies=[header_policy],
        boundary="batch_357de4f7-6d0b-4e02-8cd2-6361411a9525",
    )

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    assert request.body == (
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"DELETE /container0/blob0 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: multipart/mixed; boundary=changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"DELETE /container1/blob1 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"DELETE /container2/blob2 HTTP/1.1\r\n"
        b"x-ms-date: Thu, 14 Jun 2018 16:46:54 GMT\r\n"
        b"\r\n"
        b"\r\n"
        b"--batch_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
    )


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive(http_request, mock_response):
    class ResponsePolicy(object):
        def on_response(self, request, response):
            # type: (PipelineRequest, PipelineResponse) -> None
            response.http_response.headers["x-ms-fun"] = "true"

    req0 = http_request("DELETE", "/container0/blob0")
    req1 = http_request("DELETE", "/container1/blob1")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(req0, req1, policies=[ResponsePolicy()])

    body_as_str = (
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 202 Accepted\r\n"
        "x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        "x-ms-version: 2018-11-09\r\n"
        "\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 2\r\n"
        "\r\n"
        "HTTP/1.1 404 The specified blob does not exist.\r\n"
        "x-ms-error-code: BlobNotFound\r\n"
        "x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e2852\r\n"
        "x-ms-version: 2018-11-09\r\n"
        "Content-Length: 216\r\n"
        "Content-Type: application/xml\r\n"
        "\r\n"
        '<?xml version="1.0" encoding="utf-8"?>\r\n'
        "<Error><Code>BlobNotFound</Code><Message>The specified blob does not exist.\r\n"
        "RequestId:778fdc83-801e-0000-62ff-0334671e2852\r\n"
        "Time:2018-06-14T16:46:54.6040685Z</Message></Error>\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )

    response = mock_response(
        request,
        body_as_str.encode("ascii"),
        "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed",
    )

    response = response.parts()

    assert len(response) == 2

    res0 = response[0]
    assert res0.status_code == 202
    assert res0.headers["x-ms-fun"] == "true"

    res1 = response[1]
    assert res1.status_code == 404
    assert res1.headers["x-ms-fun"] == "true"


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_empty_requests(http_request, mock_response):

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed()

    body_as_bytes = (
        b"--batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"\r\n"
        b"HTTP/1.1 400 Bad Request\r\n"
        b"DataServiceVersion: 1.0;\r\n"
        b"Content-Type: application/xml;charset=utf-8\r\n"
        b"\r\n"
        b'<?xml version="1.0" encoding="utf-8"?><error xmlns="http://schemas.microsoft.com/ado/2007/08/dataservices/metadata"><code>InvalidInput</code><message xml:lang="en-US">An error occurred while processing this request.\nRequestId:1a930d9b-8002-0020-575c-d1b166000000\nTime:2023-08-17T22:44:06.8465534Z</message></error>\r\n'
        b"--batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f--\r\n"
    )

    response = mock_response(
        request,
        body_as_bytes,
        "multipart/mixed; boundary=batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f",
    )

    response = response.parts()
    assert len(response) == 1
    res0 = response[0]
    assert res0.status_code == 400
    assert res0.reason == "Bad Request"
    assert res0.headers["DataServiceVersion"] == "1.0;"
    assert res0.request.method == "POST"
    assert res0.request.url == "http://account.blob.core.windows.net/?comp=batch"


@pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
def test_raise_for_status_bad_response(mock_response):
    response = mock_response(request=None, body=None, content_type=None)
    response.status_code = 400
    with pytest.raises(HttpResponseError):
        response.raise_for_status()


@pytest.mark.parametrize("mock_response", MOCK_RESPONSES)
def test_raise_for_status_good_response(mock_response):
    response = mock_response(request=None, body=None, content_type=None)
    response.status_code = 200
    response.raise_for_status()


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_one_changeset(http_request, mock_response):

    changeset = http_request(None, None)
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset)

    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 202 Accepted\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 202 Accepted\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 2

    res0 = parts[0]
    assert res0.status_code == 202


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_empty_changeset(http_request, mock_response):

    changeset = http_request(None, None)
    changeset.set_multipart_mixed()
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset)

    body_as_bytes = (
        b"--batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f\r\n"
        b"Content-Type: multipart/mixed; boundary=changesetresponse_390b0b55-6892-4fce-8427-001ca15662f5\r\n"
        b"\r\n"
        b"--changesetresponse_390b0b55-6892-4fce-8427-001ca15662f5\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"\r\n"
        b"HTTP/1.1 400 Bad Request\r\n"
        b"DataServiceVersion: 1.0;\r\n"
        b"Content-Type: application/xml;charset=utf-8\r\n"
        b"\r\n"
        b'<?xml version="1.0" encoding="utf-8"?><error xmlns="http://schemas.microsoft.com/ado/2007/08/dataservices/metadata"><code>InvalidInput</code><message xml:lang="en-US">An error occurred while processing this request.\nRequestId:1a930d9b-8002-0020-575c-d1b166000000\nTime:2023-08-17T22:44:06.8465534Z</message></error>\r\n'
        b"--changesetresponse_390b0b55-6892-4fce-8427-001ca15662f5--\r\n"
        b"--batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f"
    )
    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 1
    res0 = parts[0]
    assert res0.status_code == 400
    assert res0.reason == "Bad Request"
    assert "DataServiceVersion" in res0.headers
    assert res0.request.method == "POST"
    assert res0.request.url == "http://account.blob.core.windows.net/?comp=batch"

    # Test against other HTTP verbs to see if http.client.HttpResponse has any concerns.
    changeset = http_request("PATCH", "https://foo.com")
    changeset.set_multipart_mixed()
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset)
    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f"
    )
    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 1
    res0 = parts[0]
    assert res0.status_code == 400
    assert res0.reason == "Bad Request"
    assert "DataServiceVersion" in res0.headers
    assert res0.request.method == "POST"
    assert res0.request.url == "http://account.blob.core.windows.net/?comp=batch"

    changeset = http_request("DELETE", None)
    changeset.set_multipart_mixed()
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset)
    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f"
    )
    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 1
    res0 = parts[0]
    assert res0.status_code == 400
    assert res0.reason == "Bad Request"
    assert "DataServiceVersion" in res0.headers
    assert res0.request.method == "POST"
    assert res0.request.url == "http://account.blob.core.windows.net/?comp=batch"

    changeset = http_request("HEAD", None)
    changeset.set_multipart_mixed()
    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset)
    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_b1e4a276-83db-40e9-b21f-f5bc7f7f905f"
    )
    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 1
    res0 = parts[0]
    assert res0.status_code == 400
    assert res0.reason == "Bad Request"
    assert "DataServiceVersion" in res0.headers
    assert res0.request.method == "POST"
    assert res0.request.url == "http://account.blob.core.windows.net/?comp=batch"


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_multiple_changesets(http_request, mock_response):

    changeset1 = http_request(None, None)
    changeset1.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")
    )
    changeset2 = http_request(None, None)
    changeset2.set_multipart_mixed(
        http_request("DELETE", "/container2/blob2"), http_request("DELETE", "/container3/blob3")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset1, changeset2)
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314"\r\n'
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 3\r\n"
        b"\r\n"
        b"HTTP/1.1 409\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_8b9e487e-a353-4dcb-a6f4-0688191e0314--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 4
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404
    assert parts[3].status_code == 409


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_combination_changeset_first(http_request, mock_response):

    changeset = http_request(None, None)
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), http_request("DELETE", "/container1/blob1")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(changeset, http_request("DELETE", "/container2/blob2"))
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 3
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_combination_changeset_middle(http_request, mock_response):

    changeset = http_request(None, None)
    changeset.set_multipart_mixed(http_request("DELETE", "/container1/blob1"))

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(
        http_request("DELETE", "/container0/blob0"), changeset, http_request("DELETE", "/container2/blob2")
    )
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 3
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_combination_changeset_last(http_request, mock_response):

    changeset = http_request(None, None)
    changeset.set_multipart_mixed(
        http_request("DELETE", "/container1/blob1"), http_request("DELETE", "/container2/blob2")
    )

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(http_request("DELETE", "/container0/blob0"), changeset)

    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 2\r\n"
        b"\r\n"
        b"HTTP/1.1 200\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        b'Content-Type: multipart/mixed; boundary="changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525"\r\n'
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 0\r\n"
        b"\r\n"
        b"HTTP/1.1 202\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525\r\n"
        b"Content-Type: application/http\r\n"
        b"Content-Transfer-Encoding: binary\r\n"
        b"Content-ID: 1\r\n"
        b"\r\n"
        b"HTTP/1.1 404\r\n"
        b"x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        b"x-ms-version: 2018-11-09\r\n"
        b"\r\n"
        b"\r\n"
        b"--changeset_357de4f7-6d0b-4e02-8cd2-6361411a9525--\r\n"
        b"\r\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--\r\n"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    parts = []
    for part in response.parts():
        parts.append(part)
    assert len(parts) == 3
    assert parts[0].status_code == 200
    assert parts[1].status_code == 202
    assert parts[2].status_code == 404


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_multipart_receive_with_bom(http_request, mock_response):

    req0 = http_request("DELETE", "/container0/blob0")

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(req0)
    body_as_bytes = (
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\n"
        b"Content-Type: application/http\n"
        b"Content-Transfer-Encoding: binary\n"
        b"Content-ID: 0\n"
        b"\r\n"
        b"HTTP/1.1 400 One of the request inputs is not valid.\r\n"
        b"Content-Length: 220\r\n"
        b"Content-Type: application/xml\r\n"
        b"Server: Windows-Azure-Blob/1.0\r\n"
        b"\r\n"
        b'\xef\xbb\xbf<?xml version="1.0" encoding="utf-8"?>\n<Error><Code>InvalidInput</Code><Message>One'
        b"of the request inputs is not valid.\nRequestId:5f3f9f2f-e01e-00cc-6eb1-6d00b5000000\nTime:2019-09-17T23:44:07.4671860Z</Message></Error>\n"
        b"--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )

    response = mock_response(
        request, body_as_bytes, "multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed"
    )

    response = response.parts()
    assert len(response) == 1

    res0 = response[0]
    assert res0.status_code == 400
    assert res0.body().startswith(b"\xef\xbb\xbf")


@pytest.mark.parametrize("http_request,mock_response", request_and_responses_product(MOCK_RESPONSES))
def test_recursive_multipart_receive(http_request, mock_response):
    req0 = http_request("DELETE", "/container0/blob0")
    internal_req0 = http_request("DELETE", "/container0/blob0")
    req0.set_multipart_mixed(internal_req0)

    request = http_request("POST", "http://account.blob.core.windows.net/?comp=batch")
    request.set_multipart_mixed(req0)
    internal_body_as_str = (
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 400 Accepted\r\n"
        "x-ms-request-id: 778fdc83-801e-0000-62ff-0334671e284f\r\n"
        "x-ms-version: 2018-11-09\r\n"
        "\r\n"
        "--batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed--"
    )

    body_as_str = (
        "--batchresponse_8d5f5bcd-2cb5-44bb-91b5-e9a722e68cb6\r\n"
        "Content-Type: application/http\r\n"
        "Content-ID: 0\r\n"
        "\r\n"
        "HTTP/1.1 202 Accepted\r\n"
        "Content-Type: multipart/mixed; boundary=batchresponse_66925647-d0cb-4109-b6d3-28efe3e1e5ed\r\n"
        "\r\n"
        "{}"
        "--batchresponse_8d5f5bcd-2cb5-44bb-91b5-e9a722e68cb6--"
    ).format(internal_body_as_str)

    response = mock_response(
        request,
        body_as_str.encode("ascii"),
        "multipart/mixed; boundary=batchresponse_8d5f5bcd-2cb5-44bb-91b5-e9a722e68cb6",
    )

    response = response.parts()
    assert len(response) == 1

    res0 = response[0]
    assert res0.status_code == 202

    internal_response = res0.parts()
    assert len(internal_response) == 1

    internal_response0 = internal_response[0]
    assert internal_response0.status_code == 400


def test_close_unopened_transport():
    transport = RequestsTransport()
    transport.close()


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_timeout(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    with caplog.at_level(logging.WARNING, logger="azure.core.pipeline.transport"):
        with Pipeline(transport) as pipeline:
            pipeline.run(request, connection_timeout=100)

    assert "Tuple timeout setting is deprecated" not in caplog.text


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_tuple_timeout(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    with caplog.at_level(logging.WARNING, logger="azure.core.pipeline.transport"):
        with Pipeline(transport) as pipeline:
            pipeline.run(request, connection_timeout=(100, 100))

    assert "Tuple timeout setting is deprecated" in caplog.text


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_conflict_timeout(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    with pytest.raises(ValueError):
        with Pipeline(transport) as pipeline:
            pipeline.run(request, connection_timeout=(100, 100), read_timeout=100)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_already_close_with_with(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    with Pipeline(transport) as pipeline:
        pipeline.run(request)

    # This is now closed, new requests should fail
    with pytest.raises(ValueError) as err:
        transport.send(request)
    assert "HTTP transport has already been closed." in str(err)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_already_close_manually(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    transport.send(request)
    transport.close()

    # This is now closed, new requests should fail
    with pytest.raises(ValueError) as err:
        transport.send(request)
    assert "HTTP transport has already been closed." in str(err)


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_close_too_soon_works_fine(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", "http://localhost:{}/basic/string".format(port))

    transport.close()  # Never opened, should work fine
    result = transport.send(request)

    assert result  # No exception is good enough here


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_requests_timeout_response(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", f"http://localhost:{port}/basic/string")

    with mock.patch.object(UrllibConnection, "getresponse", side_effect=SocketTimeout) as mock_method:
        with pytest.raises(ServiceResponseTimeoutError) as err:
            transport.send(request, read_timeout=0.0001)

        with pytest.raises(ServiceResponseError) as err:
            transport.send(request, read_timeout=0.0001)

        stream_request = http_request("GET", f"http://localhost:{port}/streams/basic")
        with pytest.raises(ServiceResponseTimeoutError) as err:
            transport.send(stream_request, stream=True, read_timeout=0.0001)

    stream_resp = transport.send(stream_request, stream=True)
    with mock.patch.object(UrllibResponse, "_handle_chunk", side_effect=SocketTimeout) as mock_method:
        with pytest.raises(ServiceResponseTimeoutError) as err:
            try:
                # current HttpResponse
                stream_resp.read()
            except AttributeError:
                # legacy HttpResponse
                b"".join(stream_resp.stream_download(None))


@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
def test_requests_timeout_request(caplog, port, http_request):
    transport = RequestsTransport()

    request = http_request("GET", f"http://localhost:{port}/basic/string")

    with mock.patch.object(urllib_connection, "create_connection", side_effect=SocketTimeout) as mock_method:
        with pytest.raises(ServiceRequestTimeoutError) as err:
            transport.send(request, connection_timeout=0.0001)

        with pytest.raises(ServiceRequestTimeoutError) as err:
            transport.send(request, connection_timeout=0.0001)

        stream_request = http_request("GET", f"http://localhost:{port}/streams/basic")
        with pytest.raises(ServiceRequestTimeoutError) as err:
            transport.send(stream_request, stream=True, connection_timeout=0.0001)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import sys
from requests import Response
import pytest
from azure.core.rest._requests_basic import RestRequestsTransportResponse

# NOTE: These tests are heavily inspired from the httpx test suite: https://github.com/encode/httpx/tree/master/tests
# Thank you httpx for your wonderful tests!
from azure.core.rest import HttpRequest


@pytest.fixture
def get_request_headers():
    def _get_request_headers(header_value):
        request = HttpRequest(method="GET", url="http://example.org", headers=header_value)
        return request.headers

    return _get_request_headers


# flask returns these response headers, which we don't really need for these following tests
RESPONSE_HEADERS_TO_IGNORE = [
    "Connection",
    "Content-Type",
    "Content-Length",
    "Server",
    "Date",
]


@pytest.fixture
def get_response_headers(client):
    def _get_response_headers(request):
        response = client.send_request(request)
        response.raise_for_status
        for header in RESPONSE_HEADERS_TO_IGNORE:
            response.headers.pop(header, None)
        return response.headers

    return _get_response_headers


def test_headers_request(get_request_headers):
    h = get_request_headers({"a": "123", "b": "789"})
    assert h["A"] == "123"
    assert h["B"] == "789"


def test_headers_response(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    assert "a" in h
    assert "A" in h
    assert "b" in h
    assert "B" in h
    assert "c" not in h
    assert h["a"] == "123, 456"
    assert h["A"] == "123, 456"
    assert h.get("a") == "123, 456"
    assert h.get("A") == "123, 456"
    assert h.get("nope") is None
    assert h.get("nope", default="default") is "default"
    assert h.get("nope", default=None) is None
    assert h.get("nope", default=[]) == []
    assert list(h) == ["a", "b"]

    assert set(h.keys()) == set(["a", "b"])
    assert list(h.values()) == ["123, 456", "789"]
    assert list(h.items()) == [("a", "123, 456"), ("b", "789")]
    assert list(h) == ["a", "b"]
    assert dict(h) == {"a": "123, 456", "b": "789"}


def test_headers_response_keys(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    # basically want to make sure this behaves like dict {"a": "123, 456", "b": "789"}
    ref_dict = {"a": "123, 456", "b": "789"}
    assert set(h.keys()) == set(ref_dict.keys())
    assert repr(h.keys()) == "KeysView({'a': '123, 456', 'b': '789'})"
    assert "a" in h.keys()
    assert "b" in h.keys()
    assert set(h.keys()) == set(ref_dict.keys())


def test_headers_response_keys_mutability(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    # test mutability
    before_mutation_keys = h.keys()
    h["c"] = "000"
    assert "c" in before_mutation_keys


def test_headers_response_values(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    # basically want to make sure this behaves like dict {"a": "123, 456", "b": "789"}
    ref_dict = {"a": "123, 456", "b": "789"}
    assert set(h.values()) == set(ref_dict.values())
    assert repr(h.values()) == "ValuesView({'a': '123, 456', 'b': '789'})"
    assert "123, 456" in h.values()
    assert "789" in h.values()
    assert set(h.values()) == set(ref_dict.values())


def test_headers_response_values_mutability(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    # test mutability
    before_mutation_values = h.values()
    h["c"] = "000"
    assert "000" in before_mutation_values


def test_headers_response_items(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    # basically want to make sure this behaves like dict {"a": "123, 456", "b": "789"}
    ref_dict = {"a": "123, 456", "b": "789"}
    assert set(h.items()) == set(ref_dict.items())
    assert repr(h.items()) == "ItemsView({'a': '123, 456', 'b': '789'})"
    assert ("a", "123, 456") in h.items()
    assert not ("a", "123, 456", "123, 456") in h.items()
    assert not {"a": "blah", "123, 456": "blah"} in h.items()
    assert ("A", "123, 456") in h.items()
    assert ("b", "789") in h.items()
    assert ("B", "789") in h.items()
    assert set(h.items()) == set(ref_dict.items())


def test_headers_response_items_mutability(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers"))
    # test mutability
    before_mutation_items = h.items()
    h["c"] = "000"
    assert ("c", "000") in before_mutation_items


def test_header_mutations(get_request_headers, get_response_headers):
    def _headers_check(h):
        assert dict(h) == {}
        h["a"] = "1"
        assert dict(h) == {"a": "1"}
        h["a"] = "2"
        assert dict(h) == {"a": "2"}
        h.setdefault("a", "3")
        assert dict(h) == {"a": "2"}
        h.setdefault("b", "4")
        assert dict(h) == {"a": "2", "b": "4"}
        del h["a"]
        assert dict(h) == {"b": "4"}

    _headers_check(get_request_headers({}))
    _headers_check(get_response_headers(HttpRequest("GET", "/headers/empty")))


def test_copy_headers_method(get_request_headers, get_response_headers):
    def _header_check(h):
        headers_copy = h.copy()
        assert h == headers_copy
        assert h is not headers_copy

    _header_check(
        get_request_headers(
            {
                "lowercase-header": "lowercase",
                "ALLCAPS-HEADER": "ALLCAPS",
                "CamelCase-Header": "camelCase",
            }
        )
    )
    _header_check(get_response_headers(HttpRequest("GET", "/headers/case-insensitive")))


def test_headers_insert_retains_ordering(get_request_headers, get_response_headers):
    def _header_check(h):
        h["b"] = "123"
        if sys.version_info >= (3, 6):
            assert list(h.values()) == ["a", "123", "c"]
        else:
            assert set(h.values()) == set(["a", "123", "c"])

    _header_check(get_request_headers({"a": "a", "b": "b", "c": "c"}))
    _header_check(get_response_headers(HttpRequest("GET", "/headers/ordered")))


def test_headers_insert_appends_if_new(get_request_headers, get_response_headers):
    def _headers_check(h):
        h["d"] = "123"
        if sys.version_info >= (3, 6):
            assert list(h.values()) == ["lowercase", "ALLCAPS", "camelCase", "123"]
        else:
            assert set(list(h.values())) == set(["lowercase", "ALLCAPS", "camelCase", "123"])

    _headers_check(
        get_request_headers(
            {
                "lowercase-header": "lowercase",
                "ALLCAPS-HEADER": "ALLCAPS",
                "CamelCase-Header": "camelCase",
            }
        )
    )
    _headers_check(get_response_headers(HttpRequest("GET", "/headers/case-insensitive")))


def test_headers_insert_removes_all_existing(get_request_headers, get_response_headers):
    def _headers_check(h):
        h["a"] = "789"
        assert dict(h) == {"a": "789", "b": "789"}

    _headers_check(get_request_headers([("a", "123"), ("a", "456"), ("b", "789")]))
    _headers_check(get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers")))


def test_headers_delete_removes_all_existing(get_request_headers, get_response_headers):
    def _headers_check(h):
        del h["a"]
        assert dict(h) == {"b": "789"}

    _headers_check(get_request_headers([("a", "123"), ("a", "456"), ("b", "789")]))
    _headers_check(get_response_headers(HttpRequest("GET", "/headers/duplicate/numbers")))


def test_headers_not_override():
    request = HttpRequest(
        "PUT",
        "http://example.org",
        json={"hello": "world"},
        headers={"Content-Length": "5000", "Content-Type": "application/my-content-type"},
    )
    assert request.headers["Content-Length"] == "5000"
    assert request.headers["Content-Type"] == "application/my-content-type"


def test_headers_case_insensitive(get_request_headers, get_response_headers):
    def _headers_check(h):
        assert (
            h["lowercase-header"]
            == h["LOWERCASE-HEADER"]
            == h["Lowercase-Header"]
            == h["lOwErCasE-HeADer"]
            == "lowercase"
        )
        assert h["allcaps-header"] == h["ALLCAPS-HEADER"] == h["Allcaps-Header"] == h["AlLCapS-HeADer"] == "ALLCAPS"
        assert (
            h["camelcase-header"]
            == h["CAMELCASE-HEADER"]
            == h["CamelCase-Header"]
            == h["cAMeLCaSE-hEadER"]
            == "camelCase"
        )

    _headers_check(
        get_request_headers(
            {
                "lowercase-header": "lowercase",
                "ALLCAPS-HEADER": "ALLCAPS",
                "CamelCase-Header": "camelCase",
            }
        )
    )
    _headers_check(get_response_headers(HttpRequest("GET", "/headers/case-insensitive")))


def test_multiple_headers_duplicate_case_insensitive(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/case-insensitive"))
    assert h["Duplicate-Header"] == h["duplicate-header"] == h["DupLicAte-HeaDER"] == "one, two, three"


def test_multiple_headers_commas(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/commas"))
    assert h["Set-Cookie"] == "a,  b, c"


def test_update(get_response_headers):
    h = get_response_headers(HttpRequest("GET", "/headers/duplicate/commas"))
    assert h["Set-Cookie"] == "a,  b, c"
    h.update({"Set-Cookie": "override", "new-key": "new-value"})
    assert h["Set-Cookie"] == "override"
    assert h["new-key"] == "new-value"



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
Example: Using truststore.SSLContext with Azure SDK for Python (Sync and Async)

This sample demonstrates how to use `truststore.SSLContext` directly to leverage
operating system certificate stores when working with Azure SDK for Python.

Unlike `truststore.inject_into_ssl()` which globally patches the ssl module,
this approach gives you more control over which connections use system certificates
by creating custom transport sessions with a truststore-based SSLContext.

Requirements:
    pip install truststore azure-identity azure-storage-blob aiohttp requests
USAGE:
    python example_truststore_ssl_context.py
"""

import ssl
import asyncio
import truststore
import requests
import aiohttp
from requests.adapters import HTTPAdapter

from azure.identity import DefaultAzureCredential
from azure.identity.aio import DefaultAzureCredential as AsyncDefaultAzureCredential
from azure.storage.blob import BlobServiceClient
from azure.storage.blob.aio import BlobServiceClient as AsyncBlobServiceClient
from azure.core.pipeline.transport import RequestsTransport, AioHttpTransport


# =============================================================================
# Helper: Custom HTTPAdapter for requests
# =============================================================================


class TruststoreHTTPAdapter(HTTPAdapter):
    """Custom HTTPAdapter that uses truststore.SSLContext for SSL verification."""

    def __init__(self, ssl_context, **kwargs):
        self._ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = self._ssl_context
        return super().init_poolmanager(*args, **kwargs)


# =============================================================================
# Synchronous Example
# =============================================================================


def sync_blob_storage_with_ssl_context():
    """Synchronous Azure Blob Storage using truststore.SSLContext directly.

    This approach gives you more control over the SSL context configuration
    without globally injecting truststore into ssl module.
    """

    account_url = "https://<your-storage-account>.blob.core.windows.net"

    # Create SSLContext using truststore (uses system certificate stores)
    ssl_context = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    # Create a session with truststore adapter
    session = requests.Session()
    adapter = TruststoreHTTPAdapter(ssl_context)
    session.mount("https://", adapter)

    # Create transport with the custom session
    transport = RequestsTransport(session=session, session_owner=False)

    with transport:
        credential = DefaultAzureCredential()
        blob_service_client = BlobServiceClient(
            account_url=account_url,
            credential=credential,
            transport=transport,
        )

        print("=== Sync Blob Storage with SSLContext ===")
        for container in blob_service_client.list_containers():
            print(f"Container: {container['name']}")

        blob_service_client.close()
        credential.close()

    session.close()  # Close the session manually since session_owner=False


# =============================================================================
# Asynchronous Example
# =============================================================================


async def async_blob_storage_with_ssl_context():
    """Asynchronous Azure Blob Storage using truststore.SSLContext directly.

    This approach gives you more control over the SSL context configuration
    without globally injecting truststore into ssl module.
    """

    account_url = "https://<your-storage-account>.blob.core.windows.net"

    # Create SSLContext using truststore (uses system certificate stores)
    ssl_context = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)

    # Create a TCPConnector with our truststore SSLContext
    connector = aiohttp.TCPConnector(ssl=ssl_context)
    session = aiohttp.ClientSession(connector=connector)

    # Create transport with the custom session
    transport = AioHttpTransport(session=session, session_owner=False)

    async with transport:
        credential = AsyncDefaultAzureCredential()
        blob_service_client = AsyncBlobServiceClient(
            account_url=account_url,
            credential=credential,
            transport=transport,
        )

        print("=== Async Blob Storage with SSLContext ===")
        async for container in blob_service_client.list_containers():
            print(f"Container: {container['name']}")

        await blob_service_client.close()
        await credential.close()

    await session.close()  # Close the session manually since session_owner=False


if __name__ == "__main__":
    sync_blob_storage_with_ssl_context()
    asyncio.run(async_blob_storage_with_ssl_context())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from flask import (
    request,
    jsonify as flask_jsonify,
)
from urllib.parse import urlparse, urlunparse
from .structures import CaseInsensitiveDict
import json

ENV_HEADERS = (
    "X-Varnish",
    "X-Request-Start",
    "X-Heroku-Queue-Depth",
    "X-Real-Ip",
    "X-Forwarded-Proto",
    "X-Forwarded-Protocol",
    "X-Forwarded-Ssl",
    "X-Heroku-Queue-Wait-Time",
    "X-Forwarded-For",
    "X-Heroku-Dynos-In-Use",
    "X-Forwarded-Protocol",
    "X-Forwarded-Port",
    "X-Request-Id",
    "Via",
    "Total-Route-Time",
    "Connect-Time",
)


def assert_with_message(param_name, expected_value, actual_value):
    assert expected_value == actual_value, "Expected '{}' to be '{}', got '{}'".format(
        param_name, expected_value, actual_value
    )


def jsonify(*args, **kwargs):
    response = flask_jsonify(*args, **kwargs)
    if not response.data.endswith(b"\n"):
        response.data += b"\n"
    return response


def get_url(request):
    """
    Since we might be hosted behind a proxy, we need to check the
    X-Forwarded-Proto, X-Forwarded-Protocol, or X-Forwarded-SSL headers
    to find out what protocol was used to access us.
    """
    protocol = request.headers.get("X-Forwarded-Proto") or request.headers.get("X-Forwarded-Protocol")
    if protocol is None and request.headers.get("X-Forwarded-Ssl") == "on":
        protocol = "https"
    if protocol is None:
        return request.url
    url = list(urlparse(request.url))
    url[0] = protocol
    return urlunparse(url)


def get_files():
    """Returns files dict from request context."""

    files = {}

    for k, v in request.files.items():
        content_type = request.files[k].content_type or "application/octet-stream"
        val = json_safe(v.read(), content_type)
        if files.get(k):
            if not isinstance(files[k], list):
                files[k] = [files[k]]
            files[k].append(val)
        else:
            files[k] = val

    return files


def get_headers(hide_env=True):
    """Returns headers dict from request context."""

    headers = dict(request.headers.items())

    if hide_env and ("show_env" not in request.args):
        for key in ENV_HEADERS:
            try:
                del headers[key]
            except KeyError:
                pass

    return CaseInsensitiveDict(headers.items())


def semiflatten(multi):
    """Convert a MultiDict into a regular dict. If there are more than one value
    for a key, the result will have a list of values for the key. Otherwise it
    will have the plain value."""
    if multi:
        result = multi.to_dict(flat=False)
        for k, v in result.items():
            if len(v) == 1:
                result[k] = v[0]
        return result
    else:
        return multi


def json_safe(string, content_type="application/octet-stream"):
    """Returns JSON-safe version of `string`.

    If `string` is a Unicode string or a valid UTF-8, it is returned unmodified,
    as it can safely be encoded to JSON string.

    If `string` contains raw/binary data, it is Base64-encoded, formatted and
    returned according to "data" URL scheme (RFC2397). Since JSON is not
    suitable for binary data, some additional encoding was necessary; "data"
    URL scheme was chosen for its simplicity.
    """
    try:
        string = string.decode("utf-8")
        json.dumps(string)
        return string
    except (ValueError, TypeError):
        return b"".join([b"data:", content_type.encode("utf-8"), b";base64,", base64.b64encode(string)]).decode("utf-8")


def get_dict(*keys, **extras):
    """Returns request dict of given keys."""

    _keys = ("url", "args", "form", "data", "origin", "headers", "files", "json", "method")

    assert all(map(_keys.__contains__, keys))
    data = request.data
    form = semiflatten(request.form)

    try:
        _json = json.loads(data.decode("utf-8"))
    except (ValueError, TypeError):
        _json = None

    d = dict(
        url=get_url(request),
        args=semiflatten(request.args),
        form=form,
        data=json_safe(data),
        origin=request.headers.get("X-Forwarded-For", request.remote_addr),
        headers=get_headers(),
        files=get_files(),
        json=_json,
        method=request.method,
    )

    out_d = {}

    for key in keys:
        out_d[key] = d.get(key)

    out_d.update(extras)

    return out_d


def get_base_url(request):
    return "http://" + request.host


__all__ = ["assert_with_message", "get_dict", "jsonify"]



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Tests for the distributed tracing policy in an async pipeline."""
import pytest

from azure.core.pipeline import AsyncPipeline
from azure.core.pipeline.policies import AsyncRetryPolicy, DistributedTracingPolicy
from azure.core.pipeline.transport import (
    HttpResponse,
    AsyncHttpTransport,
)

from tracing_common import FakeSpan
from utils import HTTP_REQUESTS


class MockTransport(AsyncHttpTransport):
    def __init__(self):
        self._count = 0

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def close(self):
        pass

    async def open(self):
        pass

    async def send(self, request, **kwargs):
        self._count += 1
        response = HttpResponse(request, None)
        response.status_code = 429
        return response


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_span_retry_attributes_plugin_tracing(tracing_implementation, http_request):
    """Test that the retry count is added to the span attributes with plugin tracing."""

    http_request = http_request("GET", "http://localhost/")
    retry_policy = AsyncRetryPolicy(retry_total=2)
    distributed_tracing_policy = DistributedTracingPolicy()
    transport = MockTransport()

    with FakeSpan(name="parent") as root_span:
        pipeline = AsyncPipeline(transport, [retry_policy, distributed_tracing_policy])
        await pipeline.run(http_request)

    assert transport._count == 3
    assert len(root_span.children) == 3
    assert root_span.children[0].attributes.get("http.request.resend_count") is None
    assert root_span.children[1].attributes.get("http.request.resend_count") == 1
    assert root_span.children[2].attributes.get("http.request.resend_count") == 2


@pytest.mark.asyncio
@pytest.mark.parametrize("http_request", HTTP_REQUESTS)
async def test_span_retry_attributes_native_tracing(tracing_helper, http_request):
    """Test that the retry count is added to the span attributes with native tracing."""
    http_request = http_request("GET", "http://localhost/")
    retry_policy = AsyncRetryPolicy(retry_total=2)
    distributed_tracing_policy = DistributedTracingPolicy()
    transport = MockTransport()

    with tracing_helper.tracer.start_as_current_span("Root") as root_span:
        policies = [retry_policy, distributed_tracing_policy]
        pipeline = AsyncPipeline(transport, policies=policies)
        await pipeline.run(http_request)

    assert transport._count == 3

    finished_spans = tracing_helper.exporter.get_finished_spans()
    parent_context = root_span.get_span_context()
    assert len(finished_spans) == 4
    assert finished_spans[0].parent == parent_context
    assert finished_spans[0].attributes.get(distributed_tracing_policy._HTTP_RESEND_COUNT) is None
    assert finished_spans[0].attributes.get(distributed_tracing_policy._URL_FULL) == "http://localhost/"

    assert finished_spans[1].parent == parent_context
    assert finished_spans[1].attributes.get(distributed_tracing_policy._HTTP_RESEND_COUNT) == 1
    assert finished_spans[1].attributes.get(distributed_tracing_policy._URL_FULL) == "http://localhost/"

    assert finished_spans[2].parent == parent_context
    assert finished_spans[2].attributes.get(distributed_tracing_policy._HTTP_RESEND_COUNT) == 2
    assert finished_spans[2].attributes.get(distributed_tracing_policy._URL_FULL) == "http://localhost/"



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import pytest
import os

from azure.storage.blob import BlobServiceClient
from opentelemetry.trace import SpanKind, StatusCode
from opentelemetry.sdk.trace import ReadableSpan
from devtools_testutils import get_credential


class TestTracing:

    @pytest.mark.live_test_only
    def test_blob_service_client_tracing(self, tracing_helper):
        """Test that tracing captures client method calls and HTTP requests.

        This test validates that the distributed tracing functionality properly
        captures spans when using the Blob service client, ensuring parent-child
        relationships are maintained and attributes are correctly populated.
        """
        account_url = os.environ.get("AZURE_STORAGE_BLOB_ENDPOINT")
        if not account_url:
            pytest.skip("AZURE_STORAGE_BLOB_ENDPOINT environment variable is not set.")

        client = BlobServiceClient(account_url=account_url, credential=get_credential())

        with tracing_helper.tracer.start_as_current_span(name="root") as parent:
            client.get_service_properties()

        spans = tracing_helper.exporter.get_finished_spans()
        span_names_list = [span.name for span in spans]
        # Check that last 3 spans are the expected ones (auth-related ones may vary)
        assert span_names_list[-3:] == ["GET", "BlobServiceClient.get_service_properties", "root"]

        http_span: ReadableSpan = spans[-3]
        assert http_span.kind == SpanKind.CLIENT
        assert http_span.parent
        assert http_span.parent.span_id == spans[-2].context.span_id

        assert http_span.attributes
        assert http_span.attributes["http.request.method"] == "GET"
        assert http_span.attributes["url.full"]
        assert http_span.attributes["server.address"]
        assert http_span.attributes["http.response.status_code"] == 200
        user_agent = http_span.attributes.get("user_agent.original", "")
        assert isinstance(user_agent, str) and "storage" in user_agent

        method_span: ReadableSpan = spans[-2]
        assert method_span.kind == SpanKind.INTERNAL
        assert method_span.parent
        assert method_span.parent.span_id == spans[-1].context.span_id

    @pytest.mark.live_test_only
    def test_error_handling_with_tracing(self, tracing_helper) -> None:
        """Test that tracing properly captures error information in operations.

        This test validates that when exceptions occur during operations,
        the spans correctly capture the error information.

        :param tracing_helper: Helper fixture that provides tracing functionality
        :type tracing_helper: Any
        """
        account_name = "nonexistentaccount"
        invalid_url = f"https://{account_name}.blob.core.windows.net/"

        client = BlobServiceClient(account_url=invalid_url, credential=get_credential())

        # Expecting this operation to fail
        with tracing_helper.tracer.start_as_current_span(name="root") as parent:
            try:
                client.get_service_properties()
            except Exception:
                # We expect an exception but want to verify the spans
                pass

        spans = tracing_helper.exporter.get_finished_spans()
        span_names_list = [span.name for span in spans]
        # Check that last 3 spans are the expected ones (auth-related ones may vary)
        assert span_names_list[-3:] == ["GET", "BlobServiceClient.get_service_properties", "root"]

        http_span: ReadableSpan = spans[-3]
        assert http_span.kind == SpanKind.CLIENT
        assert http_span.attributes
        assert http_span.attributes["error.type"] == "azure.core.exceptions.ServiceRequestError"
        assert http_span.status.status_code == StatusCode.ERROR



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import uuid
from time import time
from wsgiref.handlers import format_date_time

from azure.core.rest import HttpRequest
from azure.core.exceptions import (
    HttpResponseError,
    map_error,
)
from ._test_base import _TableTest


class UpdateEntityJSONTest(_TableTest):
    partition_key = str(uuid.uuid4())
    row_key = str(uuid.uuid4())

    def __init__(self, arguments):
        super().__init__(arguments)
        # base entity
        self.base_entity = self.get_base_entity(
            UpdateEntityJSONTest.partition_key, UpdateEntityJSONTest.row_key, self.args.size
        )
        self.url = f"{self.account_endpoint}{self.table_name}(PartitionKey='{UpdateEntityJSONTest.partition_key}',RowKey='{UpdateEntityJSONTest.row_key}')"

    async def global_setup(self):
        await super().global_setup()
        # create entity to be updated
        await self.async_table_client.create_entity(self.base_entity)

    def run_sync(self):
        current_time = format_date_time(time())
        request = HttpRequest(
            method="PUT",
            url=self.url,
            params={},
            headers={
                "x-ms-version": self.api_version,
                "DataServiceVersion": self.data_service_version,
                "Content-Type": "application/json",
                "Accept": "application/json",
                "If-Match": "*",
                "x-ms-date": current_time,
            },
            json=self.base_entity,
            content=None,
        )
        response = self.pipeline_client._pipeline.run(
            request,
        ).http_response
        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=self.error_map)
            raise HttpResponseError(response=response)

    async def run_async(self):
        current_time = format_date_time(time())
        request = HttpRequest(
            method="PUT",
            url=self.url,
            params={},
            headers={
                "x-ms-version": self.api_version,
                "DataServiceVersion": self.data_service_version,
                "Content-Type": "application/json",
                "Accept": "application/json",
                "If-Match": "*",
                "x-ms-date": current_time,
            },
            json=self.base_entity,
            content=None,
        )
        response = (
            await self.async_pipeline_client._pipeline.run(
                request,
            )
        ).http_response
        if response.status_code not in [204]:
            map_error(status_code=response.status_code, response=response, error_map=self.error_map)
            raise HttpResponseError(response=response)

    async def close(self):
        await self.async_table_client.close()
        await super().close()



"""
FILE: example_async_tracing.py
DESCRIPTION:
    This sample demonstrates how to trace an async client method.

    NOTE: The following example is for library developers to demonstrate how to customize tracing for their library.
    End users should be using OpenTelemetry directly for their tracing needs.

    This sample requires the opentelemetry-sdk library to be installed.
USAGE:
    python example_tracing_async.py
"""

import asyncio
from typing import Iterable, Union, Any

from azure.core.tracing.decorator_async import distributed_trace_async
from azure.core import AsyncPipelineClient
from azure.core.rest import HttpRequest, AsyncHttpResponse
from azure.core.pipeline.policies import (
    AsyncHTTPPolicy,
    SansIOHTTPPolicy,
    HeadersPolicy,
    UserAgentPolicy,
    AsyncRetryPolicy,
    DistributedTracingPolicy,
)


class SampleAsyncClient:
    """Sample async client to demonstrate how an SDK developer would add tracing to their client."""

    _instrumentation_config = {
        "library_name": "my-library",
        "library_version": "1.0.0",
        "attributes": {"az.namespace": "Sample.Namespace"},
    }

    def __init__(self, endpoint: str) -> None:
        policies: Iterable[Union[AsyncHTTPPolicy, SansIOHTTPPolicy]] = [
            HeadersPolicy(),
            UserAgentPolicy("myuseragent"),
            AsyncRetryPolicy(),
            DistributedTracingPolicy(instrumentation_config=self._instrumentation_config),
        ]

        self._client: AsyncPipelineClient[HttpRequest, AsyncHttpResponse] = AsyncPipelineClient(
            endpoint, policies=policies
        )

    @distributed_trace_async
    async def sample_method(self, **kwargs: Any) -> AsyncHttpResponse:
        request = HttpRequest("GET", "https://paulvaneck.com")
        response = await self._client.send_request(request, **kwargs)
        return response

    async def close(self) -> None:
        await self._client.close()

    async def __aenter__(self) -> "SampleAsyncClient":
        await self._client.__aenter__()
        return self

    async def __aexit__(self, *exc_details: Any) -> None:
        await self._client.__aexit__(*exc_details)


async def sample_user():
    """Sample showing how an end user would enable tracing and configure OpenTelemetry."""
    from azure.core.settings import settings

    from opentelemetry import trace
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import ConsoleSpanExporter
    from opentelemetry.sdk.trace.export import SimpleSpanProcessor

    tracer_provider = TracerProvider()
    exporter = ConsoleSpanExporter()
    span_processor = SimpleSpanProcessor(exporter)

    tracer_provider.add_span_processor(span_processor)
    trace.set_tracer_provider(tracer_provider)
    tracer = tracer_provider.get_tracer("my-application")

    settings.tracing_enabled = True

    endpoint = "https://bing.com"
    async with SampleAsyncClient(endpoint) as client:
        with tracer.start_as_current_span(name="MyApplication"):  # type: ignore
            response = await client.sample_method(tracing_options={"attributes": {"custom_key": "custom_value"}})
            print(f"Response: {response}")


if __name__ == "__main__":
    asyncio.run(sample_user())



# --------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation. All rights reserved.
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the ""Software""), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# --------------------------------------------------------------------------
import logging
import os
import sys
import pytest
from typing import NamedTuple
from unittest.mock import patch, MagicMock

# module under test
import azure.core.settings as m
from azure.core import AzureClouds


class TestPrioritizedSetting(object):
    def test_env_var_property(self):
        ps = m.PrioritizedSetting("foo", env_var="AZURE_FOO")
        assert ps.env_var == "AZURE_FOO"

    def test_everything_unset_raises(self):
        ps = m.PrioritizedSetting("foo")
        with pytest.raises(RuntimeError):
            ps()

    def test_implicit_default(self):
        ps = m.PrioritizedSetting("foo", default=10)
        assert ps() == 10

    def test_implicit_default_converts(self):
        ps = m.PrioritizedSetting("foo", convert=int, default="10")
        assert ps() == 10

    def test_system_hook(self):
        ps = m.PrioritizedSetting("foo", system_hook=lambda: 20)
        assert ps() == 20

    def test_system_hook_converts(self):
        ps = m.PrioritizedSetting("foo", convert=int, system_hook=lambda: "20")
        assert ps() == 20

    def test_env_var(self):
        os.environ["AZURE_FOO"] = "30"
        ps = m.PrioritizedSetting("foo", env_var="AZURE_FOO")
        assert ps() == "30"
        del os.environ["AZURE_FOO"]

    def test_env_var_converts(self):
        os.environ["AZURE_FOO"] = "30"
        ps = m.PrioritizedSetting("foo", convert=int, env_var="AZURE_FOO")
        assert ps() == 30
        del os.environ["AZURE_FOO"]

    def test_user_set(self):
        ps = m.PrioritizedSetting("foo")
        ps.set_value(40)
        assert ps() == 40

    def test_user_unset(self):
        ps = m.PrioritizedSetting("foo", default=2)
        ps.set_value(40)
        assert ps() == 40
        ps.unset_value()
        assert ps() == 2

    def test_user_set_converts(self):
        ps = m.PrioritizedSetting("foo", convert=int)
        ps.set_value("40")
        assert ps() == 40

    def test_immediate(self):
        ps = m.PrioritizedSetting("foo")
        assert ps(50) == 50

    def test_immediate_converts(self):
        ps = m.PrioritizedSetting("foo", convert=int)
        assert ps("50") == 50

    def test_precedence(self):
        # 0. implicit default
        ps = m.PrioritizedSetting("foo", env_var="AZURE_FOO", convert=int, default=10)
        assert ps() == 10

        # 1. system value
        ps = m.PrioritizedSetting("foo", env_var="AZURE_FOO", convert=int, default=10, system_hook=lambda: 20)
        assert ps() == 20

        # 2. environment variable
        os.environ["AZURE_FOO"] = "30"
        assert ps() == 30

        # 3. previously user-set value
        ps.set_value(40)
        assert ps() == 40

        # 4. immediate values
        assert ps(50) == 50

        del os.environ["AZURE_FOO"]

    def test___str__(self):
        ps = m.PrioritizedSetting("foo")
        assert str(ps) == "PrioritizedSetting(%r)" % "foo"

    def test_descriptors(self):
        class FakeSettings(object):
            foo = m.PrioritizedSetting("foo", env_var="AZURE_FOO")
            bar = m.PrioritizedSetting("bar", env_var="AZURE_BAR", default=10)

        s = FakeSettings()
        assert s.foo is FakeSettings.foo

        assert s.bar() == 10
        s.bar = 20
        assert s.bar() == 20


class TestConverters(object):
    @pytest.mark.parametrize("value", ["Yes", "YES", "yes", "1", "ON", "on", "true", "True", True])
    def test_convert_bool(self, value):
        assert m.convert_bool(value)

    @pytest.mark.parametrize("value", ["No", "NO", "no", "0", "OFF", "off", "false", "False", False])
    def test_convert_bool_false(self, value):
        assert not m.convert_bool(value)

    @pytest.mark.parametrize("value", [True, False])
    def test_convert_bool_identity(self, value):
        assert m.convert_bool(value) == value

    def test_convert_bool_bad(self):
        with pytest.raises(ValueError):
            m.convert_bool("junk")

    @pytest.mark.parametrize("value", ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
    def test_convert_logging_good(self, value):
        assert m.convert_logging(value) == getattr(logging, value)

        # check lowercase works too
        assert m.convert_logging(value.lower()) == getattr(logging, value)

    @pytest.mark.parametrize("value", ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
    def test_convert_logging_identity(self, value):
        level = getattr(logging, value)
        assert m.convert_logging(level) == level

    def test_convert_logging_bad(self):
        with pytest.raises(ValueError):
            m.convert_logging("junk")

    def test_convert_azure_cloud(self):
        with pytest.raises(ValueError):
            m.convert_azure_cloud(10)

    def test_convert_tracing_impl_bad(self):
        with pytest.raises(ValueError):
            m.convert_tracing_impl("foo")

    def test_convert_tracing_impl_caching(self):
        m.convert_tracing_impl.cache_clear()
        with patch.dict(
            m._tracing_implementation_dict,
            {
                "opentelemetry": MagicMock(wraps=m._get_opentelemetry_span),
                "opencensus": MagicMock(wraps=m._get_opencensus_span),
            },
        ) as patched_dict:
            mock_method = patched_dict["opentelemetry"]
            impl1 = m.convert_tracing_impl("opentelemetry")
            impl2 = m.convert_tracing_impl("opentelemetry")
            assert impl1 is impl2
            assert mock_method.call_count == 1

            mock_method = patched_dict["opencensus"]
            impl1 = m.convert_tracing_impl("opencensus")
            impl2 = m.convert_tracing_impl("opencensus")
            assert impl1 is impl2
            assert mock_method.call_count == 1
        m.convert_tracing_impl.cache_clear()


_standard_settings = ["log_level", "tracing_enabled"]


class TestStandardSettings(object):
    @pytest.mark.parametrize("name", _standard_settings)
    def test_setting_exists(self, name):
        assert hasattr(m.settings, name)

    # XXX: This test will need to become more sophisticated if the assumption
    # settings.foo -> AZURE_FOO for env vars ever becomes invalidated.
    @pytest.mark.parametrize("name", _standard_settings)
    def test_setting_env_var(self, name):
        ps = getattr(m.settings, name)
        assert ps.env_var == "AZURE_" + name.upper()

    def test_init(self):
        assert m.settings.defaults_only is False

    def test_config(self):
        val = m.settings.config(log_level=30, tracing_enabled=True)
        assert isinstance(val, tuple)
        assert val.tracing_enabled is True
        assert val.log_level == 30
        os.environ["AZURE_LOG_LEVEL"] = "debug"
        val = m.settings.config(tracing_enabled=False)
        assert val.tracing_enabled is False
        assert val.log_level == 10

        val = m.settings.config(log_level=30, tracing_enabled=False, tracing_implementation=None)
        assert val.tracing_enabled is False
        assert val.log_level == 30
        assert val.tracing_implementation is None
        del os.environ["AZURE_LOG_LEVEL"]

        val = m.settings.config(azure_cloud=AzureClouds.AZURE_US_GOVERNMENT)
        assert val.azure_cloud == AzureClouds.AZURE_US_GOVERNMENT

    def test_defaults(self):
        val: NamedTuple = m.settings.defaults
        assert val.log_level == logging.INFO
        assert val.tracing_enabled is None
        assert val.tracing_implementation is None
        assert val.azure_cloud == AzureClouds.AZURE_PUBLIC_CLOUD

    def test_tracing_setting(self):
        mock_tracing_impl = MagicMock()

        assert m.settings.tracing_enabled() is False
        assert m.settings.tracing_implementation() is None

        m.settings.tracing_enabled = None
        assert m.settings.tracing_enabled() is False
        m.settings.tracing_implementation = mock_tracing_impl

        assert m.settings.tracing_implementation() == mock_tracing_impl
        assert m.settings.tracing_enabled() is True

        m.settings.tracing_enabled = False
        assert m.settings.tracing_enabled() is False

        m.settings.tracing_implementation = None

    def test_current(self):
        os.environ["AZURE_LOG_LEVEL"] = "debug"
        val = m.settings.current
        assert isinstance(val, tuple)
        assert val.log_level == 10
        del os.environ["AZURE_LOG_LEVEL"]
        os.environ["AZURE_CLOUD"] = "AZURE_CHINA_CLOUD"
        val = m.settings.current
        assert isinstance(val, tuple)
        assert val.azure_cloud == AzureClouds.AZURE_CHINA_CLOUD
        del os.environ["AZURE_CLOUD"]



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from json import dumps
from flask import (
    Response,
    Blueprint,
)

encoding_api = Blueprint("encoding_api", __name__)


@encoding_api.route("/latin-1", methods=["GET"])
def latin_1():
    r = Response("Latin 1: ÿ".encode("latin-1"), status=200)
    r.headers["Content-Type"] = "text/plain; charset=latin-1"
    return r


@encoding_api.route("/latin-1-with-utf-8", methods=["GET"])
def latin_1_charset_utf8():
    r = Response("Latin 1: ÿ".encode("latin-1"), status=200)
    r.headers["Content-Type"] = "text/plain; charset=utf-8"
    return r


@encoding_api.route("/no-charset", methods=["GET"])
def latin_1_no_charset():
    r = Response("Hello, world!", status=200)
    r.headers["Content-Type"] = "text/plain"
    return r


@encoding_api.route("/iso-8859-1", methods=["GET"])
def iso_8859_1():
    r = Response("Accented: Österreich".encode("iso-8859-1"), status=200)  # cspell:disable-line
    r.headers["Content-Type"] = "text/plain"
    return r


@encoding_api.route("/emoji", methods=["GET"])
def emoji():
    r = Response("👩", status=200)
    return r


@encoding_api.route("/emoji-family-skin-tone-modifier", methods=["GET"])
def emoji_family_skin_tone_modifier():
    r = Response("👩🏻‍👩🏽‍👧🏾‍👦🏿 SSN: 859-98-0987", status=200)
    return r


@encoding_api.route("/korean", methods=["GET"])
def korean():
    r = Response("아가", status=200)
    return r


@encoding_api.route("/json", methods=["GET"])
def json():
    data = {"greeting": "hello", "recipient": "world"}
    content = dumps(data).encode("utf-16")
    r = Response(content, status=200)
    r.headers["Content-Type"] = "application/json; charset=utf-16"
    return r


@encoding_api.route("/invalid-codec-name", methods=["GET"])
def invalid_codec_name():
    r = Response("おはようございます。".encode("utf-8"), status=200)
    r.headers["Content-Type"] = "text/plain; charset=invalid-codec-name"
    return r


@encoding_api.route("/no-charset", methods=["GET"])
def no_charset():
    r = Response("Hello, world!", status=200)
    r.headers["Content-Type"] = "text/plain"
    return r


@encoding_api.route("/gzip", methods=["GET"])
def gzip_content_encoding():
    r = Response(
        b"\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\n\xcbH\xcd\xc9\xc9W(\xcf/\xcaI\x01\x00\x85\x11J\r\x0b\x00\x00\x00",
        status=200,
    )
    r.headers["Content-Type"] = "text/plain"
    r.headers["Content-Encoding"] = "gzip"
    return r


@encoding_api.route("/deflate", methods=["GET"])
def deflate_content_encoding():
    r = Response(b"\xcb\xc8T(\xc9H-J\x05\x00", status=200)
    r.headers["Content-Type"] = "text/plain"
    r.headers["Content-Encoding"] = "deflate"
    return r
